@@ -10,7 +10,7 @@
 # http://www.wtfpl.net/ for more details.
 
 package Chess::Plisco::Engine::Tree;
-$Chess::Plisco::Engine::Tree::VERSION = 'v1.0.0';
+$Chess::Plisco::Engine::Tree::VERSION = 'v1.0.1';
 use strict;
 use integer;
 
@@ -68,6 +68,8 @@ sub new {
 		$position->[CP_POS_REVERSIBLE_CLOCK] = @$signatures - 1;
 	}
 
+	no integer;
+
 	my @killers = map { [] } 0 .. MAX_PLY - 1;
 	my $self = {
 		position => $position,
@@ -80,6 +82,12 @@ sub new {
 		book_depth => $options{book_depth},
 		killers => \@killers,
 		cutoff_moves => [[], []], # History heuristic, one slot for each side.
+		average_score => -INF,
+		previous_average_score => INF,
+		iter_scores => [],
+		previous_time_reduction => 0.85,
+		move_efforts => {},
+		total_best_move_changes => 0,
 	};
 
 	bless $self, $class;
@@ -138,7 +146,7 @@ sub checkTime {
 	} else {
 		use integer;
 
-		my $allocated = $self->{allocated_time};
+		my $allocated = $self->{maximum};
 		my $eta = $allocated - $elapsed;
 		if ($eta < SAFETY_MARGIN) {
 			die "PLISCO_ABORTED\n";
@@ -161,11 +169,11 @@ sub checkTime {
 		#
 		# We can expect to process $eta * $nps / 1000 nodes in the remaining
 		# time. In order to play safe, should the performance suddenly drop,
-		# we divide that number by 4. A division by 1000 is roughly a
+		# we divide that number by 8. A division by 1000 is roughly a
 		# right shift of 10, a division by 4 a right-shift by 2. We can
-		# therefore just right-shift the product by 12.
+		# therefore just right-shift the product by 13.
 		if (!($was_ponder && !$self->{ponder})) {
-			my $dyn_nodes = ($eta * $nps) >> 12;
+			my $dyn_nodes = ($eta * $nps) >> 13;
 
 			my $nodes_to_go = ($max_nodes < $dyn_nodes) ? $max_nodes : $dyn_nodes;
 			$self->{nodes_to_tc} = $nodes + $nodes_to_go;
@@ -272,7 +280,7 @@ sub alphabeta {
 		my $hex_sig = sprintf '%016x', $signature;
 		$self->indent($ply, "TT probe $hex_sig \@depth $depth, alpha = $alpha, beta = $beta");
 	}
-	my $tt_value = $tt->probe($signature, $depth, $alpha, $beta, \$tt_move);
+	my $tt_value = $tt->probe($signature, $ply, $depth, $alpha, $beta, \$tt_move);
 
 	if (DEBUG) {
 		if ($tt_move) {
@@ -432,7 +440,7 @@ sub alphabeta {
 		my @line;
 		$position->move($move, 1);
 		$signatures->[$signature_slot] = $position->[CP_POS_SIGNATURE];
-		++$self->{nodes};
+		my $nodes_before = $self->{nodes}++;
 		$self->printCurrentMove($depth, $move, $legal) if $print_current_move;
 		my $score;
 		if (DEBUG) {
@@ -468,6 +476,12 @@ sub alphabeta {
 		if (DEBUG) {
 			pop @{$self->{line}};
 		}
+		if ($ply == 1) {
+			$self->{average_score} =
+				$self->{average_score} != -INF ? ($score + $self->{average_score}) >> 1 : 3.75 * $score;
+			$self->{move_efforts}->{$move} += $self->{nodes} - $nodes_before;
+			++$self->{total_best_move_changes} if $score > $best_value && $score > $alpha && $legal > 1;
+		}
 		if ($score > $best_value) {
 			$best_value = $score;
 			$best_move = $move;
@@ -530,14 +544,17 @@ sub alphabeta {
 		}
 	}
 
+	my $hash_value = $best_value;
 	if (!$legal) {
 		# Mate or stalemate.
 		if (!$position->inCheck) {
-			$best_value = DRAW;
+			$hash_value = $best_value = DRAW;
 		} else {
-			#$alpha = MATE + $self->{depth} - $depth + 1;
+			$hash_value = MATE;
 			$best_value = MATE + $ply;
 		}
+		$tt_type = Chess::Plisco::Engine::TranspositionTable::TT_SCORE_EXACT();
+
 		if (DEBUG) {
 			$self->indent($ply, "mate/stalemate, score: $best_value");
 		}
@@ -558,7 +575,7 @@ sub alphabeta {
 		}
 	}
 
-	$tt->store($signature, $depth, $tt_type, $best_value, $best_move)
+	$tt->store($signature, $depth, $tt_type, $hash_value, $best_move)
 		if !($is_null_window && $tt_type == Chess::Plisco::Engine::TranspositionTable::TT_SCORE_EXACT());
 
 	return $best_value;
@@ -597,7 +614,7 @@ sub quiesce {
 		my $hex_sig = sprintf '%016x', $signature;
 		$self->indent($ply, "quiescence TT probe $hex_sig \@depth 0, alpha = $alpha, beta = $beta");
 	}
-	my $tt_value = $tt->probe($signature, 0, $alpha, $beta, \$tt_move);
+	my $tt_value = $tt->probe($signature, $ply, 0, $alpha, $beta, \$tt_move);
 	if (DEBUG) {
 		if ($tt_move) {
 			my $cn = $position->moveCoordinateNotation($tt_move);
@@ -754,12 +771,20 @@ sub rootSearch {
 	my $alpha = -INF;
 	my $beta = +INF;
 
+	my $last_best_move;
+	my $last_best_move_depth = 0;
+
 	if (DEBUG) {
 		my $fen = $position->toFEN;
 		$self->debug("Searching $fen");
 	}
 	eval {
 		while (++$depth <= $max_depth) {
+			no integer;
+
+			# Age out instability.
+			$self->{total_best_move_changes} /= 2;
+
 			my @lower_windows = (-50, -100, -INF);
 			my @upper_windows = (50, 100, +INF);
 
@@ -784,6 +809,65 @@ sub rootSearch {
 				$beta = $score + ASPIRATION_WINDOW;
 				redo;
 			}
+
+			if ($self->{use_time_management}) {
+				my $iter_idx = ($depth - 1) & 3;
+
+				no integer;
+
+				# The constants are taken from Stockfish and they use their own units
+				# which seem to be 3.5-4.0 centipawns.
+				my $best_value = 3.75 * $score;
+
+				my $falling_eval = (11.85 + 2.24 * ($self->{previous_average_score} - $best_value)
+					+ 0.93 * ($self->{iter_scores}->[$iter_idx] - $best_value)) / 100.0;
+				$falling_eval = ($falling_eval) < (0.57) ? (0.57) : ($falling_eval) > (1.70) ? (1.70) : ($falling_eval);
+				if ($line[0] != $last_best_move) {
+					$last_best_move_depth = $depth;
+				}
+
+				# Idea: Try to move the center based on the average completed
+				# search depth. But the average search depth should probably
+				# be aged out.
+				# my $depth_scale = $self->{avg_completed_depth} / 40.0; # SF baseline
+				# my $center = $last_best_move_depth + 12.15 * $depth_scale;
+				my $k = 0.51; # FIXME! Lower that to 0.25?
+				my $center = $last_best_move_depth + 12.15; # Divide by 4.25?
+
+				my $time_reduction = 0.66 + 0.85 / (0.98 + exp(-$k * ($depth - $center)));
+				my $reduction = (1.43 + $self->{previous_time_reduction}) / (2.28 * $time_reduction);
+				my $best_move_instability = 1.02 + 2.14 * $self->{total_best_move_changes};
+				my $nodes_effort = $self->{move_efforts}->{$line[0]} * 100000 / (((1) > ($self->{nodes})) ? (1) : ($self->{nodes}));
+
+				# The original value is 93340. But we will not reach that
+				# with only safe prunings.
+				my $high_best_move_effort = $nodes_effort >= 30000 ? 0.76 : 1.0;
+
+				my $total_time = $self->{optimum} * $falling_eval * $reduction
+					* $best_move_instability * $high_best_move_effort;
+
+				my $elapsed = 1000 * tv_interval($self->{start_time});
+
+				if ($elapsed > ((($total_time) < ($self->{maximum})) ? ($total_time) : ($self->{maximum}))) {
+					if (!$self->{ponder}) {
+						last;
+					}
+				} elsif (!$self->{ponder}) {
+					# Stockfish sets the threshold to half of the total time.
+					# We use less.
+					#
+					# Current best: 0.25
+					# Values 0.5, 0.375, 0.3125 all failed SPRTs.
+					if ($elapsed > $total_time >> 2) {
+						last;
+					}
+				}
+
+				$self->{previous_average_score} = $self->{average_score};
+				$self->{previous_time_reduction} = $time_reduction;
+				$last_best_move = $line[0];
+				$self->{iter_scores}->[$iter_idx] = $best_value;
+			}
 		}
 	};
 	if ($@) {
@@ -792,10 +876,10 @@ sub rootSearch {
 		}
 	}
 
-	if ($self->{allocated_time} && !$self->{ponderhit}) {
+	if ($self->{maximum} && !$self->{ponderhit}) {
 		my $elapsed = 1000 * tv_interval($self->{start_time});
-		if ($elapsed > $self->{allocated_time}) {
-			$self->{info}->(__"Error: used $elapsed ms instead of $self->{allocated_time} ms.");
+		if ($elapsed > $self->{maximum}) {
+			$self->{info}->(__"Error: used $elapsed ms instead of $self->{maximum} ms.");
 		}
 	}
 
@@ -803,6 +887,22 @@ sub rootSearch {
 }
 
 
+sub outputMoveEfforts {
+	my ($self) = @_;
+
+	my $sum = 0;
+	my $all_nodes = $self->{nodes} || 1;
+	foreach my $move (sort { $self->{move_efforts}->{$b} <=> $self->{move_efforts}->{$a} } keys %{$self->{move_efforts}}) {
+		my $san = $self->{position}->SAN($move);
+		my $nodes = $self->{move_efforts}->{$move};
+		$sum += $nodes;
+		my $effort = int(0.5 + (100000 * $nodes / $all_nodes));
+		$self->{info}->("Move effort $san: $effort ($nodes nodes)");
+	}
+
+	$self->{info}->("Move effort sum: $sum");
+}
+
 sub printCurrentMove {
 	my ($self, $depth, $move, $moveno) = @_;
 
@@ -812,6 +912,7 @@ sub printCurrentMove {
 	my $cn = $position->moveCoordinateNotation($move);
 	my $elapsed = int(1000 * tv_interval($self->{start_time}));
 
+	$moveno += 1;
 	$self->{info}->("depth $depth currmove $cn currmovenumber $moveno"
 		. " time $elapsed");
 }
@@ -846,7 +947,7 @@ sub think {
 	$self->{tt_hits} = 0;
 
 	if ($self->{debug}) {
-		$self->{info}->("allocated time: $self->{allocated_time}");
+		$self->{info}->("allocated time: $self->{optimum} .. $self->{maximum}");
 	}
 
 	$self->rootSearch(\@line);
@@ -897,7 +998,7 @@ sub getPonderMove {
 		$self->debug("probing transposition table for ponder move");
 	}
 	# We're not interested in the value.
-	$tt->probe($signature, MAX_PLY + 1, 0, 0, \$tt_move);
+	$tt->probe($signature, 1, MAX_PLY + 1, 0, 0, \$tt_move);
 
 	if (DEBUG) {
 		if ($tt_move) {
