@@ -14,13 +14,13 @@
 ## no critic (TestingAndDebugging::RequireUseStrict)
 
 (do {
-	# FIXME! Pass the color in to the macro!
-	my $my_color = cp_pos_to_move($p);
-	my $her_pieces = $p->[CP_POS_WHITE_PIECES + !$my_color];
-	my $occupancy = ($p->[CP_POS_WHITE_PIECES + $my_color] | $her_pieces) & ~(1 << $from);
+	# FIXME! Pass the colour in to the macro!
+	my $my_colour = cp_pos_to_move($p);
+	my $her_pieces = $p->[CP_POS_WHITE_PIECES + !$my_colour];
+	my $occupancy = ($p->[CP_POS_WHITE_PIECES + $my_colour] | $her_pieces) & ~(1 << $from);
 	my $queens = cp_pos_queens($p);
 	$her_pieces
-		& (($pawn_masks[$my_color]->[2]->[$to] & cp_pos_pawns($p))
+		& (($pawn_masks[$my_colour]->[2]->[$to] & cp_pos_pawns($p))
 			| ($knight_attack_masks[$to] & cp_pos_knights($p))
 			| ($king_attack_masks[$to] & cp_pos_kings($p))
 			| (cp_mm_bmagic($to, $occupancy) & ($queens | cp_pos_bishops($p)))
