@@ -1,6 +1,6 @@
 #! /bin/false
 
-# Copyright (C) 2021 Guido Flohr <guido.flohr@cantanea.com>,
+# Copyright (C) 2021-2025 Guido Flohr <guido.flohr@cantanea.com>,
 # all rights reserved.
 
 # This program is free software. It comes without any warranty, to
@@ -14,12 +14,13 @@
 ## no critic (TestingAndDebugging::RequireUseStrict)
 
 while ($target_mask) {
-
-	my $base_move = $b | cp_bitboard_count_trailing_zbits $t;
+	my $to = cp_bitboard_count_trailing_zbits $t;
+	my $base_move = $b | ($to << 15) | ($board[$to] << 3);
+	# It is important for move ordering to sort from good to bad promotions.
 	push @m,
-		$b | (CP_QUEEN << 12),
-		$b | (CP_ROOK << 12),
-		$b | (CP_BISHOP << 12),
-		$b | (CP_KNIGHT << 12);
+		$b | (5 << 6), # Queen
+		$b | (4 << 6), # Rook
+		$b | (3 << 6), # Bishop
+		$b | (2 << 6); # Knight
 	$target_mask = cp_bitboard_clear_least_set $target_mask;
 }
